#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import pathlib
from typing import Set, Iterator, Optional, Union, Mapping

import inflection
import typic

from iambic import parse, ast


class Corpus:

    PATH: pathlib.Path = pathlib.Path(__file__).parent
    CACHE_SIZE: int = 1000

    def __init__(self):
        self._corpus: Mapping[str, pathlib.Path] = {x.name: x for x in self}
        self.__hits: int = 0

    def _maybe_reset(self):
        if self.__hits > 1000:
            self.__getitem__.cache_clear()
            self.__hits = 0

    @staticmethod
    def _get_names(path: pathlib.Path) -> Set[str]:
        names = {path.name}
        for child in path.iterdir():
            if child.suffix == ".md":
                names.add(inflection.parameterize(child.stem))
        return names

    def __iter__(self) -> Iterator[pathlib.Path]:
        stack = {self.PATH}
        while stack:
            d = stack.pop()
            for f in d.iterdir():
                if f.is_dir():
                    stack.add(f)
                elif f.is_file() and f.suffix == ".md":
                    yield f

    @typic.fastcachedmethod
    def __getitem__(self, item: str):
        self._maybe_reset()
        if not isinstance(item, str):
            raise ValueError(
                f"Corpus keys must be type str, provided {type(item)}: {item!r}"
            )
        if item in self._corpus:
            self.__hits += 1
            return self._corpus[item].read_text()
        seen = set()
        item = inflection.parameterize(item)
        for k, v in self._corpus.items():
            if item in k:
                self.__hits += 1
                return v.read_text()
        raise KeyError(
            f"Couldn't locate a play with key {item!r}. Available keys are: {(*seen,)}"
        )

    def get(
        self, item: str, default: str = None, *, parsed: bool = False, tree: bool = True
    ) -> Optional[Union[str, ast.Play, ast.Index]]:
        try:
            text = self[item]
            return parse.text(text, tree=tree) if parsed else text
        except KeyError:
            return default


corpus = Corpus()
get = corpus.get



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# flake8: noqa
import enum

import typic

from .base import *
from .node import *
from .index import *


class InputType(str, enum.Enum):
    HTML = "html"
    MD = "markdown"
    DATA = "data"


typic.register(deserializer=node_coercer, check=isnodetype)
typic.resolve()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['docs',
 'docs.plays',
 'iambic',
 'iambic.ast',
 'iambic.bin',
 'iambic.parse',
 'iambic.render']

package_data = \
{'': ['*'],
 'docs': ['img/*', 'usage/*'],
 'docs.plays': ['comedies/*',
                'histories/*',
                'problems/*',
                'roman/*',
                'romances/*',
                'tragedies/*']}

install_requires = \
['html2text>=2020.1.16,<2021.0.0',
 'tablib>=1.1.0,<2.0.0',
 'typical>=2.0.0,<3.0.0',
 'ujson>=2.0,<3.0',
 'yattag>=1.13,<2.0']

setup_kwargs = {
    'name': 'iambic',
    'version': '1.0.1',
    'description': 'Data extraction and rendering library for Shakespearean text.',
    'long_description': "iambic: Data extraction and rendering library for Shakespearean text. :scroll: \n==============================================================================\n[![image](https://img.shields.io/pypi/v/iambic.svg)](https://pypi.org/project/iambic/)\n[![image](https://img.shields.io/pypi/l/iambic.svg)](https://pypi.org/project/iambic/)\n[![image](https://img.shields.io/pypi/pyversions/iambic.svg)](https://pypi.org/project/iambic/)\n[![image](https://img.shields.io/github/languages/code-size/seandstewart/iambic.svg?style=flat)](https://github.com/seandstewart/iambic)\n[![image](https://img.shields.io/travis/seandstewart/iambic.svg)](https://travis-ci.org/seandstewart/iambic)\n[![codecov](https://codecov.io/gh/seandstewart/iambic/branch/master/graph/badge.svg)](https://codecov.io/gh/seandstewart/iambic)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![Netlify Status](https://api.netlify.com/api/v1/badges/91ace14b-e26e-4026-ac5c-3e5640f2910f/deploy-status)](https://app.netlify.com/sites/iambic/deploys)\n\nLet computers do the hard work for you! `iambic` provides:\n1. The most accurate method for counting lines automatically.\n2. Automatically track which characters are speaking in any scene.\n3. Deterministic, repeatable results, with the ability to store your\n   data as JSON with strictly defined schema for passing over the wire\n   or storing locally or in a NoSQL database between runtimes.\n\n\n## Installation\n\nIn order to install the latest version, simply `pip3 install\n-U iambic`.\n\nThis library requires Python 3.7 or greater.\n\n\n## What is it?\n`iambic` was originally envisioned as a tool for translating\nShakespearean text into actionable information, i.e.:\n1. How many lines are in this particular play?\n2. How many lines does a given character speak in this play?\n3. Which characters speak, in which scenes and acts?\n\nAs a result of the implementation, this tool can be applied \nto any body of text which adhere's to its parsing syntax.\n\n\n## The Schema\nThe full schema specification has been written in JSON\nSchema 7.0 and can be found\n[here](schema.json)\n\n## Documentation\n\nThe full documentation is available at\n[iambic.seandstewart.io](https://iambic.seandstewart.io)\n\n\n## How to Contribute\n1.  Check for open issues or open a fresh issue to start a \n    discussion around a feature idea or a bug.\n2.  Create a branch on Github for your issue or fork\n    [the repository](https://github.com/seandstewart/iambic)\n    on GitHub to start making your changes to the **master**\n    branch.\n3.  Write a test which shows that the bug was fixed or that \n    the feature works as expected.\n4.  Send a pull request and bug the maintainer until it gets\n     merged and published. :)\n",
    'author': 'Sean Stewart',
    'author_email': 'sean_stewart@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/seandstewart/iambic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import dataclasses
import functools
from typing import ClassVar, Optional, Union, Tuple, Any, Mapping, Type, List, TypeVar

import inflection
import typic

from iambic import roman
from .base import NodeType, NodeMixin


__all__ = (
    "Act",
    "Scene",
    "Prologue",
    "Epilogue",
    "Intermission",
    "Persona",
    "Entrance",
    "Exit",
    "Action",
    "Direction",
    "Dialogue",
    "Speech",
    "ResolvedNode",
    "ChildNode",
    "NodeTree",
    "Play",
    "Metadata",
    "GenericNode",
    "node_coercer",
    "isnodetype",
)


T = TypeVar("T")


@typic.klass(unsafe_hash=True)
class Act(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.ACT)
    index: int
    text: str
    num: int

    @typic.cached_property
    def id(self):
        return inflection.parameterize(f"act-{self.col}")

    @typic.cached_property
    def col(self):
        return roman.numeral(self.num)

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        return cls(index=node.index, text=node.match_text, num=num)


@typic.klass(unsafe_hash=True)
class Scene(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.SCENE)
    index: int
    text: str
    num: int
    act: str
    setting: Optional[str]

    @typic.cached_property
    def id(self) -> str:
        return inflection.parameterize(f"{self.act}-scene-{roman.numeral(self.num)}")

    @typic.cached_property
    def col(self) -> str:
        if self.act in {NodeType.PROL, NodeType.EPIL}:
            pre = self.act[0].upper()
        else:
            pre = self.act.split("-")[-1].upper()
        return f"{pre}.{roman.numeral(self.num).lower()}"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        setting = " ".join(node.pieces[2:]) if len(node.pieces) > 2 else None
        parent = node.parent or node.act or node.scene
        parent = parent.resolve() if hasattr(parent, "resolve") else parent
        return cls(
            index=node.index, text=node.match_text, num=num, act=parent, setting=setting
        )


@typic.klass(unsafe_hash=True)
class Prologue(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.PROL)
    index: int
    text: str
    setting: Optional[str]
    act: str = None

    @typic.cached_property
    def id(self):
        return f"{self.act}-prologue" if self.act else "prologue"

    @typic.cached_property
    def col(self):
        pre = f"{self.act.split('-')[-1].upper()}." if self.act else ""
        return f"{pre}P"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        setting = " ".join(node.pieces[1:]) if len(node.pieces) > 1 else None
        return cls(
            index=node.index, text=node.match_text, setting=setting, act=node.parent
        )


@typic.klass(unsafe_hash=True)
class Epilogue(Prologue):
    type: NodeType = dataclasses.field(init=False, default=NodeType.EPIL)

    @typic.cached_property
    def id(self):
        return f"{self.act}-epilogue" if self.act else "epilogue"

    @typic.cached_property
    def col(self):
        pre = f"{self.act.split('-')[-1].upper()}." if self.act else ""
        return f"{pre}E"


@typic.klass(unsafe_hash=True)
class Intermission(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.INTER)
    index: int
    text: str
    act: str

    @typic.cached_property
    def id(self):
        return f"intermission"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(index=node.index, text=node.match_text, act=node.parent)

    @typic.cached_property
    def col(self):
        return "INT"


@typic.klass(unsafe_hash=True)
class Persona(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.PERS)
    index: int
    text: str
    name: str
    short: str = None

    def __eq__(self, other) -> bool:
        return other.name == self.name if hasattr(other, "name") else False

    @typic.cached_property
    def id(self):
        return inflection.parameterize(self.name)

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(
            index=node.index,
            text=node.match_text,
            name=inflection.titleize(node.match_text),
        )


@typic.klass(unsafe_hash=True)
class Entrance(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.ENTER)
    index: int
    text: str
    scene: str
    personae: Tuple[str, ...] = dataclasses.field(default_factory=tuple)

    @typic.cached_property
    def id(self):
        return f"{self.scene}-entrance-{self.index}"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(index=node.index, text=node.match_text, scene=node.parent)


@typic.klass(unsafe_hash=True)
class Exit(Entrance):
    type: NodeType = dataclasses.field(init=False, default=NodeType.EXIT)

    @typic.cached_property
    def id(self):
        return f"{self.scene}-exit-{self.index}"


@typic.klass(unsafe_hash=True)
class Action(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.ACTION)
    action: str
    persona: str
    scene: str
    index: int

    @typic.cached_property
    def id(self):
        return f"{self.scene}-{self.persona}-action-{self.index}"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(
            action=node.match_text,
            persona=node.parent,
            scene=node.scene,
            index=node.index,
        )


@typic.klass(unsafe_hash=True)
class Direction(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.DIR)
    action: str
    scene: str
    index: int
    stop: bool = True

    @typic.cached_property
    def id(self):
        return f"{self.scene}-direction-{self.index}"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(action=node.match_text, scene=node.parent, index=node.index)


@typic.klass(unsafe_hash=True)
class Dialogue(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.DIAL)
    line: str
    persona: str
    scene: str
    index: int
    lineno: int
    linepart: int = 0

    @typic.cached_property
    def id(self):
        return f"{self.persona}-dialogue-{self.lineno}-{self.linepart}"

    @classmethod
    def from_node(cls: Type[T], node: "GenericNode") -> T:
        return cls(
            node.match_text,
            node.parent,
            node.scene,
            index=node.index,
            lineno=node.lineno,
            linepart=node.linepart,
        )


@typic.klass(unsafe_hash=True, delay=True)
class Speech(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.SPCH)
    persona: str
    scene: str
    speech: Tuple[Union[Dialogue, Action, Direction], ...]
    index: int

    @typic.cached_property
    def linerange(self) -> Tuple[int, int]:
        linenos = sorted([x.lineno for x in self.speech if hasattr(x, "lineno")])
        return linenos[0], linenos[-1]

    @typic.cached_property
    def num_lines(self) -> int:
        x, y = self.linerange
        # line count starts at 1
        # a range of 1 - 1 is 1 line
        return y - (x - 1)

    @typic.cached_property
    def id(self) -> str:
        return (
            f"{self.scene}-{self.persona}-speech-{'-'.join(map(str, self.linerange))}"
        )


ResolvedNode = Union[
    Act,
    Scene,
    Prologue,
    Epilogue,
    Persona,
    Entrance,
    Exit,
    Action,
    Direction,
    Dialogue,
    Speech,
    Intermission,
]

ChildNode = ResolvedNode


@typic.klass(unsafe_hash=True, delay=True)
class NodeTree(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.TREE)
    node: ResolvedNode
    children: Tuple[ChildNode, ...] = dataclasses.field(default_factory=tuple)
    personae: Tuple[str, ...] = dataclasses.field(default_factory=tuple)

    @typic.cached_property
    def cols(self) -> Tuple[Any, ...]:
        return tuple(
            getattr(x, "node", x).col
            for x in self.children
            if isinstance(x, (NodeTree, Prologue))
        )

    @typic.cached_property
    def id(self) -> str:
        return f"{self.node.id}-tree"


ChildNode = Union[
    Act,
    Scene,
    Prologue,
    Epilogue,
    Persona,
    Entrance,
    Exit,
    Action,
    Direction,
    Dialogue,
    Speech,
    Intermission,
    NodeTree,
]


@typic.klass(unsafe_hash=True)
class Metadata(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.META)
    rights: str = "Creative Commons Non-Commercial Share Alike 3.0"
    language: str = "en-GB-emodeng"
    publisher: str = "Published w/ ❤️ using iambic - https://pypi.org/project/iambic"
    title: str = None
    subtitle: str = None
    edition: int = 1
    author: str = "William Shakespeare"
    editors: Tuple[str, ...] = dataclasses.field(default_factory=tuple)
    tags: Tuple[str, ...] = dataclasses.field(default_factory=tuple)

    @functools.lru_cache(1)
    def asmeta(self):  # pragma: nocover
        dikt = {
            "creator": [{"type": "author", "text": self.author}],
            "contributor": [{"type": "editor", "text": "MIT"}],
            "rights": self.rights,
            "language": self.language,
            "publisher": self.publisher,
            "subject": ["Shakespeare"],
        }
        if self.title:
            dikt["title"] = [
                {"type": "main", "text": self.title},
                {"type": "edition", "text": self.edition},
            ]
            if self.subtitle:
                dikt["title"].append({"type": "subtitle", "text": self.subtitle})
        dikt["contributor"].extend(
            ({"type": "editor", "text": x} for x in self.editors or [])
        )
        dikt["subject"].extend(self.tags or [])

        return dikt


@typic.klass(unsafe_hash=True, delay=True)
class Play(NodeMixin):
    type: NodeType = dataclasses.field(init=False, default=NodeType.PLAY)
    children: Tuple[NodeTree, ...] = dataclasses.field(default_factory=tuple)
    personae: Tuple[Persona, ...] = dataclasses.field(default_factory=tuple)
    meta: Metadata = dataclasses.field(default_factory=Metadata)

    @typic.cached_property
    def id(self) -> str:
        return inflection.parameterize(f"{self.meta.title}-play")


@dataclasses.dataclass(unsafe_hash=True)
class GenericNode(NodeMixin):
    """The root-object of a script.

    A script ``Node`` represents a single line of text in a script.
    """

    __resolver_map__: ClassVar[
        Mapping[NodeType, Type[ResolvedNode]]
    ] = typic.FrozenDict(
        {
            NodeType.ACT: Act,
            NodeType.ACTION: Action,
            NodeType.DIAL: Dialogue,
            NodeType.DIR: Direction,
            NodeType.ENTER: Entrance,
            NodeType.EPIL: Epilogue,
            NodeType.EXIT: Exit,
            NodeType.INTER: Intermission,
            NodeType.META: Metadata,
            NodeType.PERS: Persona,
            NodeType.PLAY: Play,
            NodeType.PROL: Prologue,
            NodeType.SCENE: Scene,
            NodeType.SPCH: Speech,
            NodeType.TREE: NodeTree,
        }
    )
    # Minimum data for a Node
    type: NodeType
    text: str
    index: int
    # Additional fields which may be present
    lineno: int = None
    linepart: int = None
    # Given by text parser
    # If reading from JSON, we don't have/need this,
    # it will be provided inherently by the data-structure
    # on resolution-time.
    match: typic.FrozenDict = dataclasses.field(default_factory=typic.FrozenDict)
    parent: str = None
    act: str = None
    scene: str = None

    @typic.cached_property
    def resolved(self) -> ResolvedNode:
        """Resolve a GenericNode into a typed, "resolved" Node.

        Raises
        ------
        TypeError
            If the NodeType provided has no resolved Node mapping.
        """
        try:
            return self.__resolver_map__[self.type].from_node(self)
        except KeyError:  # pragma: nocover
            raise TypeError(
                f"Unrecognized node-type <{self.type}> for text <{self.text}>. "
                f"Valid types are: {tuple(self.__resolver_map__)}"
            )

    @typic.cached_property
    def pieces(self) -> List[str]:
        return self.match_text.split()

    @typic.cached_property
    def match_text(self) -> str:
        return self.match[self.type] if self.match else self.text


_RESOLVABLE = set(GenericNode.__resolver_map__.values())
_candidates = _RESOLVABLE.union(ChildNode.__args__)
_candidates.add(NodeTree)


def node_coercer(value: Any) -> Optional[ResolvedNode]:
    if type(value) in _RESOLVABLE or value is None:
        return value
    if isinstance(value, GenericNode):
        return value.resolved

    if not isinstance(value, Mapping):
        value = typic.coerce(value, dict)

    handler: Type[ResolvedNode] = GenericNode.__resolver_map__[value.pop("type")]
    return handler(**value)


@functools.lru_cache(maxsize=None)
def isnodetype(obj: Type) -> bool:
    is_valid = (
        obj is ResolvedNode
        or obj is GenericNode
        or obj in _candidates
        or (
            getattr(obj, "__origin__", None) is Union
            and set(getattr(obj, "__args__", set())).issubset(_candidates)
        )
    )
    return is_valid



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
from collections import defaultdict
from typing import (
    Sequence,
    DefaultDict,
    Dict,
    Type,
    Optional,
    List,
    Set,
    Union,
    ValuesView,
)

import iambic.ast.node as ast

__all__ = ("Index",)


IndexKeyType = Union[
    ast.NodeType, ast.ResolvedNode, Type[ast.ResolvedNode], ast.GenericNode, str
]
SpeechMemberType = Union[ast.Action, ast.Dialogue, ast.Direction]


class Index(Dict[str, ast.ResolvedNode]):
    """An ordered collection of nodes in a script.

    ``Parser.parse`` builds a list of :class:`GenericNode`, which can be resolved into the specific
    data-type for the :class:`NodeType` assigned.
    """

    def __init__(self, initlist: Sequence[ast.GenericNode] = None):
        initlist = initlist or []
        super().__init__(initlist)
        self.generic = {}
        self.__type_map: DefaultDict[
            ast.NodeType, Dict[str, ast.ResolvedNode]
        ] = defaultdict(dict)

    def __getitem__(self, item: IndexKeyType):
        if isinstance(item, ast.NodeType):
            return self.__type_map[item]
        return super().__getitem__(item)

    def __contains__(self, item: IndexKeyType) -> bool:
        if isinstance(item, ast.NodeType):
            return item in self.__type_map
        return super().__contains__(item)

    def get(self, k: IndexKeyType, default=None):
        default = default or {}
        if k in self:
            return self[k]
        return default

    def get_values(self, k: ast.NodeType) -> ValuesView[ast.ResolvedNode]:
        return self.get(k, default={}).values()

    def add(self, node: ast.GenericNode):
        resolved = node.resolved
        self[resolved.id] = resolved
        self.generic[resolved.id] = node
        self.__type_map[resolved.type][resolved.id] = resolved

    @property
    def personae(self) -> ValuesView[ast.Persona]:
        return self.get_values(ast.NodeType.PERS)

    @property
    def acts(self) -> ValuesView[ast.Act]:
        return self.get_values(ast.NodeType.ACT)

    @property
    def scenes(self) -> ValuesView[ast.Scene]:
        return self.get_values(ast.NodeType.SCENE)

    @property
    def prologues(self) -> ValuesView[ast.Prologue]:
        return self.get_values(ast.NodeType.PROL)

    @property
    def epilogues(self) -> ValuesView[ast.Epilogue]:
        return self.get_values(ast.NodeType.EPIL)

    @property
    def dialogue(self) -> ValuesView[ast.Dialogue]:
        return self.get_values(ast.NodeType.DIAL)

    @property
    def directions(self) -> ValuesView[ast.Direction]:
        return self.get_values(ast.NodeType.DIR)

    @property
    def actions(self) -> ValuesView[ast.Action]:
        return self.get_values(ast.NodeType.ACTION)

    @property
    def entrances(self) -> ValuesView[ast.Entrance]:
        return self.get_values(ast.NodeType.ENTER)

    @property
    def exits(self) -> ValuesView[ast.Exit]:
        return self.get_values(ast.NodeType.EXIT)

    @property
    def intermission(self) -> Optional[ast.Intermission]:
        inter = {*self.get_values(ast.NodeType.INTER)}
        return inter.pop() if inter else None

    def resolve_presence(self):
        members: List[Union[ast.Entrance, ast.Exit]] = [*self.entrances] + [*self.exits]
        personae = {x.text: x for x in self.personae}
        for member in members:
            present = tuple(y.id for x, y in personae.items() if x in member.text)
            member.personae = present

    def get_speeches(self) -> List[ast.Speech]:
        # Candidates for members of speeches.
        members: List[SpeechMemberType] = sorted(
            [*self.dialogue] + [*self.directions] + [*self.actions], key=self.node_sort
        )
        persona: Optional[ast.Persona] = None
        scene: Optional[ast.Scene] = self[members[0].scene]
        speech: List[SpeechMemberType] = []
        speeches: List[ast.Speech] = []

        for node in members:
            # If we're in a new scene, we're definitely in a new speech.
            if node.scene != scene.id:
                spch = ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
                speeches.append(spch)
                # Directions aren't directly associated to a persona,
                # so do not have the persona attr.
                # BUT they can occur within speeches, so we still have to track them all.
                persona = None if node.type == ast.NodeType.DIR else self[node.persona]
                scene, persona, speech = self[node.scene], persona, [node]
                continue
            # These puppies have a `persona` attr, so are easily associated to a speech.
            if node.type in {ast.NodeType.DIAL, ast.NodeType.ACTION}:
                # If the persona isn't set or has been reset.
                # See above for when persona can be set to null.
                if not persona:
                    persona = self[node.persona]
                    speech.append(node)
                    continue
                # Short-circuit the below if the personae match.
                if node.persona == persona.id:
                    speech.append(node)
                    continue
                # Otherwise, we have a new persona, meaning we have a new speech.
                spch = ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
                speeches.append(spch)
                persona, scene, speech = self[node.persona], self[node.scene], [node]
                continue
            # If we've set a persona and we come across a Direction,
            # we can be reasonably sure this dir should be associated to this speech.
            if node.type == ast.NodeType.DIR and persona:
                speech.append(node)
        # Once we've exhausted all members, it's possible we have one speech left.
        if persona and scene and speech:
            spch = ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
            speeches.append(spch)

        return speeches

    def filter_directions(self, speeches: Set[ast.Speech]) -> Set[ast.Direction]:
        speech_directions = set(
            y for x in speeches for y in x.speech if isinstance(y, ast.Direction)
        )
        return set(self.get(ast.Direction, {}).values()) - speech_directions

    def get_scene_trees(self, sort: bool = True) -> List[ast.NodeTree]:
        speeches = {*self.get_speeches()}
        directions = self.filter_directions(speeches)
        entrances = {*self.entrances}
        exits = {*self.exits}
        children = speeches | directions | entrances | exits
        scenes = []
        for node in {*self.scenes} | {*self.prologues} | {*self.epilogues}:
            child_nodes = set(x for x in children if x.scene == node.id)
            personae = set(
                x.persona for x in child_nodes if x.type == ast.NodeType.SPCH
            )
            if child_nodes:
                tree = ast.NodeTree(
                    node=node,
                    children=tuple(sorted(child_nodes, key=self.node_sort)),
                    personae=tuple(personae),
                )
                scenes.append(tree)
                children -= child_nodes
        if sort:
            scenes.sort(key=self.node_sort)
        return scenes

    def get_act_trees(self, scenes: List[ast.NodeTree]) -> List[ast.NodeTree]:
        scenes = {*scenes}
        acts = {x for x in scenes if x.node.act is None}
        scenes -= acts
        intermission = self.intermission
        prols = set(x for x in self.prologues if x.act is None)
        epils = set(x for x in self.epilogues if x.act is None)
        for act in prols | {*self.acts} | epils:
            children = set(x for x in scenes if x.node.act == act.id)
            if children:
                scenes -= children
                # Add the intermission if there is one.
                if intermission and intermission.act == act.id:
                    children.add(intermission)
                nodes = tuple(sorted(children, key=self.node_sort))
                acts.add(ast.NodeTree(act, nodes))
        acts = list(sorted(acts, key=self.node_sort))
        return acts

    def to_tree(self, meta: ast.Metadata = None) -> ast.Play:
        # Resolve the presence of personae in entrances/exits
        self.resolve_presence()
        # Build the Act-level trees
        scenes = self.get_scene_trees()
        # Build the Play-level trees
        acts = self.get_act_trees(scenes)
        # Put it all together
        play = ast.Play((*acts,), (*self.personae,), meta=(meta or ast.Metadata()))
        return play

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def node_sort(node):
        return getattr(node, "node", node).index
